package object_detection.types;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.*;

public class ObjectSet {

    public List<PointSet> objects;
    public ObjectSet(){
        objects = new ArrayList<>();
    }

    /**
     * This method is used to create a new object (or candidate), that will then be compared to others
     * @param pp : the points that will be used as features
     * @return : the index of this object, so we can quickly add and remove objects from list
     */
    public int makeObject(Point...pp){
        // blank pointset
        PointSet ps = new PointSet(objects.size());
        ps.addAll(pp);

        // add all points
        ps.updateReps();

        // add pointset to object list
        objects.add(ps);

        // return index of new object, which will be at end of ArrayList
        return objects.size()-1;
    }

    /**
     * We compare two objects
     * @param i : should be the already in database object
     * @param j : the new candidate object
     * @return true if objects are the same
     */
    public boolean compareObjects(int i, int j){
        // O(n), moving points
        Point[] r1 = this.objects.get(i).getPoints();
        Point[] r2 = this.objects.get(j).getPoints();

        int count = 0;

        // O(n^2)
        for (Point point : r1) {
            for (Point value : r2) {
                if (point.equals(value)) {
                    // we found point r1[x], now find next, and so on
                    count++;
                    break;
                }
            }
        }

        // in this line, if the reps are 20% similar, assume this is the same set.
        return (count) > (Math.max(r1.length, r2.length)*0.1);
    }

    /**
     * This function merges object j into object i (these are indexes in objects set)
     * @param i : current object in list
     * @param j : new object to be combined into i
     */
    public void combineObjects(int i, int j){

        // add all points of j to i
        this.objects.get(i).addAll(this.objects.get(j).getPoints());

        // redo reps of the object with new points
        this.objects.get(i).updateReps();

        // remove j
        this.objects.remove(j);
    }

    /**
     * This method updates the ObjectSet to incorporate new information from the features and bounding boxes of a new keyframe.
     *
     * @param featCsv : file containing each feature in space (x,y,z and index)
     * @param bboxCsv : file containing Bounding Boxes for each object in a keyframe
     */
    
    public List<ArrayList<Point>> processFrame(File featCsv, File bboxCsv) throws FileNotFoundException {
        // Print the name of the file being processed
        //System.out.println("Processing " + featCsv.getName() + " and " + bboxCsv.getName());

        // Read bounding boxes
        List<BoundingBox> boundingBoxes = readBoundingBoxes(bboxCsv);

        // Initialize results structure
        List<ArrayList<Point>> results = new ArrayList<>();
        for (int i = 0; i < boundingBoxes.size(); i++) {
            results.add(new ArrayList<>());
        }

        // Process feature points
        Scanner scanner = new Scanner(featCsv);
        while (scanner.hasNextLine()) {

            // get the Point and its corresponding 2D projection from the current KeyFrame
            String[] line = scanner.nextLine().split(",");
            Point2D point2D = new Point2D(Float.parseFloat(line[0]), Float.parseFloat(line[1]), Integer.parseInt(line[2]));
            Point point = new Point(Float.parseFloat(line[3]), Float.parseFloat(line[4]), Float.parseFloat(line[5]), 0);

            // Check if the point is within any bounding box
            for (int i = 0; i < boundingBoxes.size(); i++) {
                if (boundingBoxes.get(i).within(point2D)) {
                    results.get(i).add(point);
                    break; // Assuming a point belongs to only one bounding box, first come first serve
                }
            }
        }
        scanner.close();

        return results;
    }

    /**
     * This method updates the existing ObjectSet to incorporate new information from the features and bounding boxes of a new keyframe.
     * It also updates the ObjectSet to merge objects that are similar.
     * 
     * @param points : the features of the new keyframe of the form List<ArrayList<Point>>
     * 
     */
    public void updateObjectSet(List<ArrayList<Point>> points) {
        // for each BoundingBox we have all points that fall within it

        for(ArrayList<Point> singleBoxPoints : points) {
            if(singleBoxPoints.isEmpty()){
                continue;
            }
            Point[] pointArray = singleBoxPoints.toArray(new Point[0]);
            int objIdx = makeObject(pointArray);

            // Compare the new object with all existing objects
            for (int j = 0; j < this.objects.size(); j++) {
                // don't compare with yourself
                if(objIdx == j){
                    continue;
                }
                if (compareObjects(objIdx, j)) {
                    combineObjects(j, objIdx);
                    break;
                }
            }
        }
    }

    /**
     * Takes a file of bounding boxes; just another utility function
     * @param bboxCsv : File object holding info about a single csv of bounding boxes from one frame
     * @return : list of BoundingBox objects from that KeyFrame
     * @throws FileNotFoundException
     */
    private List<BoundingBox> readBoundingBoxes(File bboxCsv) throws FileNotFoundException {
        List<BoundingBox> boxes = new ArrayList<>();
        Scanner scanner = new Scanner(bboxCsv);
        scanner.nextLine(); // Skip header

        int i = 0;
        while (scanner.hasNextLine()) {
            String[] line = scanner.nextLine().split(",");

            // get all coordinate points
            Point2D tr = new Point2D(Float.parseFloat(line[2]) + Float.parseFloat(line[4]), Float.parseFloat(line[3]), 0);
            Point2D tl = new Point2D(Float.parseFloat(line[2]), Float.parseFloat(line[3]), 0);
            Point2D br = new Point2D((Float.parseFloat(line[2]) + Float.parseFloat(line[4])), Float.parseFloat(line[3]) + Float.parseFloat(line[5]), 0);
            Point2D bl = new Point2D(Float.parseFloat(line[2]), (Float.parseFloat(line[3]) + Float.parseFloat(line[5])), 0);

            // create the box accordingly
            boxes.add(new BoundingBox(tr, tl, br, bl, i, line[0]));
            i++;
        }
        scanner.close();
        return boxes;
    }

    /**
     * Just used for debugging purposes
     * @return res : String representation of the object set
     */
    @Override
    public String toString(){
        return "ObjectSet of : " + this.objects.size() + " objects";
    }
}